'''
## This file is autogenerated from macrosynergy setup.py.

It contains version information for the macrosynergy package.

short_version: 0.1.10

version: 0.1.10

full_version: 0.1.10

git_revision: 499a50af6474eaa932c572ff9aa0633865ba11c5

release: True

'''
short_version: str = '0.1.10'
version: str = '0.1.10'
full_version: str = '0.1.10'
git_revision: str = '499a50af6474eaa932c572ff9aa0633865ba11c5'
release: bool = True
if not release:
    version = full_version
